# by Matthias Buergler, 2021
import numpy as np
import pathlib
import sys
import math
import scipy
from scipy.sparse import csc_matrix, lil_matrix, tril, find


class NODE:
    def __init__ (self, x=0.0, y=0.0, z=0.0):
        self.__coords = [x, y, z]
        self.__index = None
        self.__attribute = None

    def __getitem__ (self,ind):
        return self.__coords[ind]

    def setID (self, index):
        self.__index = index

    def getID (self):
        return self.__index

    def getX (self):
        return self.__coords[0]

    def getY (self):
        return self.__coords[1]

    def getZ (self):
        return self.__coords[2]

    def getPointXY(self):
        return [self.__coords[0], self.__coords[1]]

    def getPointXYZ(self):
        return [self.__coords[0], self.__coords[1], self.__coords[2]]

    def setAttribute(self, attribute):
        self.__attribute = attribute

    def getAttribute(self):
        return self.__attribute


class EDGE:
    # CLASS of segment, defined by 2 nodes
    def __init__ (self, nodes):
        self.__name = None
        if len(nodes)==2:
            self.__nodes = nodes # list of node objects
            self.calcProperties()
        else:
            print ('to define a segment, exactly two nodes are needed!')
            return

    def __eq__ (self, other):
        out = 0
        for ni in self.__nodes:
            for nj in other.getNodes():
                if ni==nj:
                    out = out+1
        if out==len(self.__nodes):
            return True
        else:
            return False

    def __str__ (self):
        return ("(%f/%f) -> (%f/%f)"%
            (self.__nodes[0].getX(),self.__nodes[0].getY(),self.__nodes[1].getX(),self.__nodes[1].getY()))

    def getNodes (self):
        return self.__nodes

    def calcProperties (self):
        n1 = self.__nodes[0]
        n2 = self.__nodes[1]
        # calculate edge length
        self.__length = math.sqrt( math.pow(n1.getX()-n2.getX(),2) + math.pow(n1.getY()-n2.getY(),2) )
        #if (self.__length < 0.1):
        #    print 'n1 %f %f'%(n1.getX(),n1.getY())
        #    print 'n2 %f %f'%(n2.getX(),n2.getY())
        #    print 'WARNING, length is small'
        # calcualte normal vector
        #vec = np.array( [np.cross( [n2.getX()-n1.getX(), n2.getY()-n1.getY(), 0.0], [0.0, 0.0, 1.0] )] ).transpose()
        #self.__normal = np.divide( vec, self.__length )
        # calculate the orientation
        self.__orientation = [(n2.getX()-n1.getX())/self.__length, (n2.getY()-n1.getY())/self.__length]

    def getOrientation (self):
        return self.__orientation

    def setName (self, name):
        self.__name = name

    def getName (self):
        return self.__name
    
    def getLength (self):
        return self.__length

class CELL:

    def __init__ (self, node1, node2, node3):
        self.__nodes = [node1,node2,node3]
        self.__matID = None
        self._center = [0.0,0.0,0.0]

    def calcProperties (self,center='average'):
        nodes = self.getNodes()
        # calculate the element center coordinates
        x = 0.0
        y = 0.0
        N = len(nodes)
        if center=='average':
            # average of triangle nodes coordinates as interpolation point
            for nn in nodes:
                x = x+nn.getX()
                y = y+nn.getY()
            x = x/3.0
            y = y/3.0
        elif center=='outer':
            # outer circle center as interpolation point
            d =  0.0
            for ii in range(N):
                n1 = nodes[(ii-1)%N]
                n2 = nodes[ii]
                n3 = nodes[(ii+1)%N]
                dx = n3.getX()-n1.getX()
                dy = n3.getY()-n1.getY()
                d = d + n2.getX()*dy
                v = n2.getX()**2 + n2.getY()**2
                x = x + v*dy
                y = y - v*dx
            x = x / (2.0*d)
            y = y / (2.0*d)
        elif center=='inner':
            # inner circle cener as interpolation point
            lsum = 0.0
            for ii in range(N):
                n1 = nodes[(ii+1)%N]
                n2 = nodes[(ii+2)%N]
                l = math.sqrt((n2.getX()-n1.getX())**2 + (n2.getY()-n1.getY())**2)
                lsum += l
                x = x + l*nodes[ii].getX()
                y = y + l*nodes[ii].getY()
            x = x/lsum
            y = y/lsum
        elif center=='mid':
            # mid of xmax and xmin (ymax and ymin)
            xmax = max(nodes[0].getX(),max(nodes[1].getX(),nodes[2].getX()))
            xmin = min(nodes[0].getX(),min(nodes[1].getX(),nodes[2].getX()))
            ymax = max(nodes[0].getY(),max(nodes[1].getY(),nodes[2].getY()))
            ymin = min(nodes[0].getY(),min(nodes[1].getY(),nodes[2].getY()))
            x = 0.5*(xmin+xmax)
            y = 0.5*(ymin+ymax)
        #print x,y
        self._center = [x, y, 0.0]

    def getCenter (self):
        return self._center

    def setElevation (self, z):
        self._center[2] = z

    def getElevation (self):
        return self._center[2]

    #def getElementCenterCoords (self, elID):
    #   return self._elements[elID].getCenter()

    def getNodes (self):
        return self.__nodes

    def getNodeIDs (self):
        nids = []
        for ni in self.__nodes:
            nids.append(ni.getID())
        return nids

    def setMaterial (self, value):
        self.__matID = value

    def getMaterial (self):
        return self.__matID

    def calcArea (self):
        n1,n2,n3 = self.getNodes()
        return abs((n1.getX()*(n2.getY()-n3.getY()) + n2.getX()*(n3.getY()-n1.getY())+ n3.getX()*(n1.getY()-n2.getY()))/2.0)

    def calcAngles (self):
        n1,n2,n3 = self.getNodes()
        P12 = np.sqrt((n1.getX() - n2.getX())**2.0 + (n1.getY() - n2.getY())**2.0)
        P13 = np.sqrt((n1.getX() - n3.getX())**2.0 + (n1.getY() - n3.getY())**2.0)
        P23 = np.sqrt((n2.getX() - n3.getX())**2.0 + (n2.getY() - n3.getY())**2.0)
        angles = []
        angles.append(np.arccos((P12**2.0 + P13**2.0 - P23**2.0) / (2.0 * P12 * P13))/np.pi*180.0)
        angles.append(np.arccos((P12**2.0 + P23**2.0 - P13**2.0) / (2.0 * P12 * P23))/np.pi*180.0)
        angles.append(np.arccos((P23**2.0 + P13**2.0 - P12**2.0) / (2.0 * P23 * P13))/np.pi*180.0)
        return angles
        
    def isInside (self, x, y):
        whichSide = lambda pt, n1, n2: True if (pt[0] - n2.getX()) * (n1.getY() - n2.getY()) - (pt[1] - n2.getY())*(n1.getX() - n2.getX()) >= 0.0 else False
        side1 = whichSide((x, y), self.__nodes[0], self.__nodes[1])
        N = len(self.__nodes)
        for ii in range(1, N):
            side2 = whichSide((x, y), self.__nodes[ii], self.__nodes[(ii+1)%N])
            if side2 != side1:
                return False
            else:
                side1 = side2
        return True

    def interpolate (self, x, y):
        P = self.__nodes[0]
        Q = self.__nodes[1]
        R = self.__nodes[2]
        a =  -(R.getY()*Q.getZ() - P.getY()*Q.getZ() - R.getY()*P.getZ() + Q.getY()*P.getZ() + P.getY()*R.getZ() - Q.getY()*R.getZ())
        b =   P.getY()*R.getX() + Q.getY()*P.getX() + R.getY()*Q.getX() - Q.getY()*R.getX() - P.getY()*Q.getX() - R.getY()*P.getX()
        c =   Q.getZ()*R.getX() + P.getZ()*Q.getX() + R.getZ()*P.getX() - P.getZ()*R.getX() - Q.getZ()*P.getX() - Q.getX()*R.getZ()
        d = -a*P.getX() - b*P.getZ() - c*P.getY()
        if b == 0.0:
            if ((x == P.getX()) and (y == P.getY)):
                return P.getZ()
            if (x == Q.getX() and y == Q.getY()):
                return Q.getZ()
            if (x == R.getX() and y == R.getY()):
                return R.getZ()
            return 0.0
        return -(a*x+c*y+d) / b


class MESH2DM:
# by Samuel J. Peter, 2015
    def __init__ (self, filename):
        n,c = self.__read2dm(filename)
        self.__nodeDict = n
        self.__cellList = c
        self.__EdgeList = []
        self.__addEdges()

    def getMesh (self):
        return (self.__nodeDict,self.__cellList)
    
    def calcMeshStatistics(self):
        cellAreas = []
        cellAngles = []
        for ce in self.__cellList:
            cellAreas.append(ce.calcArea())
            for an in ce.calcAngles():
                cellAngles.append(an)
        edgeLengths = []
        for ed in self.__EdgeList:
            edgeLengths.append(ed.getLength())
        self.__cellAreas = cellAreas
        self.__cellAngles = cellAngles
        self.__edgeLengths = edgeLengths

    def writeMeshStatistics(self,outfile):
        nNodes =  len(self.__nodeDict)
        nCells =  len(self.__cellList)
        writer = open(outfile,'w')
        try:
            writer.write("Number of Elements,,%d\n"%(nCells))
            writer.write("Number of Nodes,,%d\n"%(nNodes))
            writer.write("Area [m^2],Min.,%.6f\n"%(min(self.__cellAreas)))
            writer.write(",Max.,%.6f\n"%(max(self.__cellAreas)))
            writer.write(",Mean.,%.6f\n"%(np.mean(self.__cellAreas)))
            writer.write(",SD.,%.6f\n"%(np.std(self.__cellAreas)))
            writer.write("Length [m],Min.,%.6f\n"%(min(self.__edgeLengths)))
            writer.write(",Max.,%.6f\n"%(max(self.__edgeLengths)))
            writer.write(",Mean.,%.6f\n"%(np.mean(self.__edgeLengths)))
            writer.write(",SD.,%.6f\n"%(np.std(self.__edgeLengths)))
            writer.write("Angle [degree],Min.,%.6f\n"%(min(self.__cellAngles)))
            writer.write(",Max.,%.6f\n"%(max(self.__cellAngles)))
            writer.write(",Mean.,%.6f\n"%(np.mean(self.__cellAngles)))
            writer.write(",SD.,%.6f\n"%(np.std(self.__cellAngles)))
        finally:
            writer.close()

    def __getMeshMaps (self,nNodes,connArr):
        # init sparse matrix to map nodes->edges
        n2e = lil_matrix((nNodes,nNodes),dtype=np.int)
        e2cDict = {} # dict map edge->cells
        edgeID = 1 # we start at 1 because 0 means 'no-data'
        for ic,ns in enumerate(connArr):
            for ii in range(len(ns)):
                n1 = ns[ii]
                n2 = ns[(ii+1)%len(ns)]
                eid = n2e[n1,n2]
                if not eid:
                    # new edge
                    n2e[n1,n2] = edgeID
                    n2e[n2,n1] = edgeID
                    e2cDict[edgeID] = [ic,None]
                    edgeID += 1
                else:
                    # existing edge
                    e2cDict[eid][1] = ic
        n1,n2,ei = find(tril(n2e))
        e2nDict = {}
        for ii in range(ei.shape[0]):
            e2nDict[ei[ii]] = [n1[ii],n2[ii]]
        return (e2nDict,e2cDict,n2e)

    def __read2dm (self,filename2dm):
        # import .2dm file, generated by SMS or QGIS
        try:
            smsFile = open(filename2dm,'r')
        except:
            print ("could not open the file '%s'!" % filename2dm)
            return
        lineCounter = 0
        elData = []
        cellList = []
        nodeDict = {}
        for line in smsFile:
            # parse line
            tokens = line.strip().split()
            if len(tokens) < 1:
                continue
            if tokens[0] in ["MESH2d", "MESH2D", "MESHNAME"]:
                continue
            # store triangle data in list
            if tokens[0] == "E3T":
                if len(tokens) < 6:
                    print ("Error found in 2dm file at cell-definition at line #%d. Cannot proceed." % lineCounter+1)
                    return
                else:
                    elData.append( [int(tokens[2]), int(tokens[3]), int(tokens[4]), int(tokens[5])] )
            # store quadrilateral data in list
            elif tokens[0] == "E4Q":
                print ("Error found in 2dm file at cell-definition at line #%d. No support for quadrilateral elements." % lineCounter+1)
                return
            # add node to mesh
            elif tokens[0] == "ND":
                if len(tokens) < 5:
                    print ("Error found in 2dm file at node-definition at line #%d. Cannot proceed." % lineCounter+1)
                    return
                else:
                    nodeID = int(tokens[1])
                    node = NODE( float(tokens[2]), float(tokens[3]), float(tokens[4]) )
                    node.setID(nodeID)
                    nodeDict[nodeID] = node
            # go to next line
            lineCounter = lineCounter + 1
        smsFile.close()
        # create the cells
        for eli in elData:
            cell = CELL( nodeDict[eli[0]],nodeDict[eli[1]],nodeDict[eli[2]] )
            cell.calcProperties()
            cell.setMaterial( eli[3] )
            cellList.append( cell )
        # return the results
        return (nodeDict,cellList)

    def __mesh2array (self):
        NN = len(self.__nodeDict)
        NC = len(self.__cellList)
        coords = np.zeros((NN,3),dtype=np.float)
        c2n = np.zeros((NC,3),dtype=np.int)
        # correct all ids because in 2dm we start from 1
        for key,ni in self.__nodeDict.iteritems():
            coords[key-1,:] = [ni.getX(),ni.getY(),ni.getZ()]
        for ii,ci in enumerate(self.__cellList):
            nodes = ci.getNodes()
            c2n[ii,:] = [nodes[0].getID()-1, nodes[1].getID()-1, nodes[2].getID()-1]
        return (coords,c2n)
    
    def __addEdges (self):
        for ce in self.__cellList:
            n1,n2,n3 = ce.getNodes()
            pairs =  [[n1,n2],[n1,n3],[n2,n3]]
            for p in pairs:
                newEdge = EDGE([p[0],p[1]])
                # do we really have a new edge?
                new = True
                for ii,e in enumerate(self.__EdgeList):
                    if e==newEdge:
                        new = False 
                if new:
                    self.__EdgeList.append( newEdge )

mesh = ['R30/BED30_1sqm.2dm',\
    'R30/BED30_5sqm.2dm',\
    'R30/BED30_10sqm.2dm',\
    'R30/BED30_20sqm.2dm',\
    'R30/BED30_100sqm.2dm',\
    'R80/BED80mod_1sqm.2dm',\
    'R80/BED80mod_5sqm.2dm',\
    'R80/BED80mod_10sqm.2dm',\
    'R80/BED80mod_20sqm.2dm',\
    'R80/BED80mod_100sqm.2dm',\
    'R150/BED150_1sqm.2dm',\
    'R150/BED150_5sqm.2dm',\
    'R150/BED150_10sqm.2dm',\
    'R150/BED150_20sqm.2dm',\
    'R150/BED150_100sqm.2dm']
path = '/media/matthias/scratch/Projekte/02_Publikationen/A_PA_Wicki_Buergler2017/TJoE/Topography/BASEMENT_Geometry_files/2D/'
for me in mesh:
    file2DM = path+me
    path = pathlib.Path(file2DM)
    m = MESH2DM(path)
    m.calcMeshStatistics()
    m.writeMeshStatistics(path.parent / (path.name.replace(path.suffix,'') + '_stats.txt'))

